@@ -39,6 +39,7 @@ import time
 import zlib
 
 from kazoo.exceptions import (
+    BadArgumentsError,
     BadVersionError,
     InvalidACLError,
     NoAuthError,
@@ -591,7 +592,19 @@ class Shell(AugumentedCmd):
 
         example:
         exists /foo
-        ZnodeStat(czxid=101, mzxid=102, ctime=1382820644375, mtime=1382820693801, version=1, cversion=0, aversion=0, ephemeralOwner=0, dataLength=6, numChildren=0, pzxid=101)
+        Stat(
+          czxid=101,
+          mzxid=102,
+          ctime=1382820644375,
+          mtime=1382820693801,
+          version=1,
+          cversion=0,
+          aversion=0,
+          ephemeralOwner=0,
+          dataLength=6,
+          numChildren=0,
+          pzxid=101
+        )
 
         # sets a watch
         exists /foo true
@@ -604,7 +617,20 @@ class Shell(AugumentedCmd):
         path = self.resolve_path(params.path)
         stat = self._zk.exists(path, **kwargs)
         if stat:
-            self.do_output(str(stat))
+            session = stat.ephemeralOwner if stat.ephemeralOwner else 0
+            self.do_output("Stat(")
+            self.do_output("  czxid=%s", stat.czxid)
+            self.do_output("  mzxid=%s", stat.mzxid)
+            self.do_output("  ctime=%s", stat.ctime)
+            self.do_output("  mtime=%s", stat.mtime)
+            self.do_output("  version=%s", stat.version)
+            self.do_output("  cversion=%s", stat.cversion)
+            self.do_output("  aversion=%s", stat.aversion)
+            self.do_output("  ephemeralOwner=0x%x", session)
+            self.do_output("  dataLength=%s", stat.dataLength)
+            self.do_output("  numChildren=%s", stat.numChildren)
+            self.do_output("  pzxid=%s", stat.pzxid)
+            self.do_output(")")
         else:
             self.do_output("Path %s doesn't exist", params.path)
 
@@ -687,7 +713,9 @@ class Shell(AugumentedCmd):
         except NotEmptyError:
             self.do_output("%s is not empty.", params.path)
         except NotReadOnlyCallError:
-            self.do_output("Not a read-only operation")
+            self.do_output("Not a read-only operation.")
+        except BadArgumentsError:
+            self.do_output("Bad arguments.")
 
     complete_rm = _complete_path
 
@@ -822,6 +850,8 @@ child_watches=%s"""
             self._zk.delete(params.path, recursive=True)
         except NotReadOnlyCallError:
             self.do_output("Not a read-only operation")
+        except BadArgumentsError:
+            self.do_output("Bad arguments.")
 
     complete_rmr = _complete_path
 
@@ -1097,6 +1127,95 @@ child_watches=%s"""
                 if repeat > 0 and i >= repeat:
                     break
 
+    @connected
+    @ensure_params(
+        Required("path"),
+        Required("hosts"),
+        BooleanOptional("recursive"),
+        BooleanOptional("reverse")
+    )
+    @check_paths_exists("path")
+    def do_ephemeral_endpoint(self, params):
+        """
+        gets the session and ip:port for an ephemeral znode
+
+        ephemeral_endpoint <path> <hosts> [recursive: bool] [reverse_lookup: bool]
+
+        where hosts is a list of hosts in the host1[:port1][,host2[:port2]],... form
+
+        examples:
+
+        ephemeral_endpoint /servers/member_0000044941 10.0.0.1,10.0.0.2,10.0.0.3
+        0xa4788b919450e6 10.3.2.12:54250 10.0.0.2:2181
+
+        """
+        if invalid_hosts(params.hosts):
+            self.do_output("List of hosts has the wrong syntax.")
+            return
+
+        stat = self._zk.exists(params.path)
+        if stat is None:
+            self.do_output("%s is gone.", params.path)
+            return
+
+        if not params.recursive and stat.ephemeralOwner == 0:
+            self.do_output("%s is not ephemeral.", params.path)
+            return
+
+        try:
+            info_by_path = self._zk.ephemerals_info(params.hosts)
+        except AugumentedClient.CmdFailed as ex:
+            self.do_output(str(ex))
+            return
+
+        def check(path, show_path, resolved):
+            info = info_by_path.get(path, None)
+            if info is None:
+                self.do_output("No session info for %s.", path)
+            else:
+                self.do_output("%s%s",
+                               "%s: " % (path) if show_path else "",
+                               info.resolved if resolved else str(info))
+
+        if not params.recursive:
+            check(params.path, False, params.reverse)
+        else:
+            for cpath, _ in self._zk.tree(params.path, 0, full_path=True):
+                check(cpath, True, params.reverse)
+
+    complete_ephemeral_endpoint = _complete_path
+
+    @connected
+    @ensure_params(Required("session"), Required("hosts"), BooleanOptional("reverse"))
+    def do_session_endpoint(self, params):
+        """
+        gets the client_ip:port and server_ip:port for the given session
+
+        session_endpoint <session> <hosts>
+
+        where hosts is a list of hosts in the host1[:port1][,host2[:port2]],... form
+
+        examples:
+
+        session_endpoint 0xa4788b919450e6 10.0.0.1,10.0.0.2,10.0.0.3
+        10.3.2.12:54250 10.0.0.2:2181
+        """
+        if invalid_hosts(params.hosts):
+            self.do_output("List of hosts has the wrong syntax.")
+            return
+
+        try:
+            info_by_id = self._zk.sessions_info(params.hosts)
+        except AugumentedClient.CmdFailed as ex:
+            self.do_output(str(ex))
+            return
+
+        info = info_by_id.get(params.session, None)
+        if info is None:
+            self.do_output("No session info for %s.", params.session)
+        else:
+            self.do_output("%s", info.resolved_endpoints if params.reverse else info.endpoints)
+
     @ensure_params(Required("hosts"))
     def do_connect(self, params):
         """
